#include "c51\cpragma.h"                /* pragma-definitions */
#pragma code
/* DATE :               VERSION :      */

/***************************************************************************/
/*                                                                         */
/* module: SIO_MOD.C                                                       */
/* autor :                                                                 */
/*                                                                         */
/* date  :                                                                 */
/*                                                                         */
/* notes: . all values assume operation at 11.0592 MHz                     */
/*                                                                         */
/***************************************************************************/

#include <intrins.h>
#include <reg52.h>

#define SIO_MOD_DEFINED

#include "c51\sio_mod.h"
#include "c51\cmacdef.h"                /* macrodefinitions */
#include "c51\ram.h"
#include "c51\cglob.h"
#include "c51\cport.h"

/***************************************************************************/
/* external                                                                */
/***************************************************************************/
#if(WATCHDOG)
extern void handle_watchdog(void);
#endif

/***************************************************************************/
/* port definitions                                                        */
/***************************************************************************/
/* sbit baud_sel1    = P1^3;  baudrate selection jumper */
/* sbit baud_sel0    = P1^4;
/* sbit parity_sel   = P1^5;  parity selection jumper */

bit not_cts;      /* = P3^2;  hardware flow control lines should be a port-pin */
bit not_rts;      /* = P3^4;  not used because software flow control */

/***************************************************************************/
/* local                                                                   */
/***************************************************************************/

static char idata tx_buf[TX_BUF_LEN];     /* transmit buffer */
static char idata rx_buf[RX_BUF_LEN];     /* receive buffer */

static bit send_xon;        /* send xon request */
static bit send_xoff;       /* send coff request */
static bit sent_xoff;       /* xoff has been send */
static bit tx_off;          /* send is disabled due to received xoff */
static bit tx_ready;        /* transmitter is ready */
static bit even_parity;     /* used mode is: 7 bit even else 8 bit none */
static bit hard_flow;       /* use hardware flow control else xon, xoff */
static bit look_ahead_used; /* look ahead character in use */

static char tx_in,          /* tx buffer input index */
            tx_out,         /* tx buffer output index */
            tx_count,       /* tx buffer character count */
            tx_delay;       /* tx delay between char's */

static char rx_in,          /* rx buffer input index */
            rx_out,         /* rx buffer output index */
            rx_count,       /* rx buffer character count */
            msg_count,      /* rx buffer message count */
            look_ahead;     /* used for ungetch() */

#ifdef AUTO_XON
static int give_xon_cnt;    /* no message timeout (auto xon) */
#endif
#ifdef RELAXED_XOFF
static int ignore_xoff_cnt; /* send timeout (ignore xoff) */
#endif

sbit ac7       = ACC^7;     /* used for acc.7 access */
sfr16 TH2L2    = 0xCC;      /* used for timer 2 int access */
sfr16 RCAP2H2L = 0xca;      /* used for rcap2 int access */

/***************************************************************************/
/* global                                                                  */
/***************************************************************************/

bit parity_err;             /* received character has a parity error */

/***************************************************************************/
/* function prototypes                                                     */
/***************************************************************************/

void sio_init(char soft_flow_control);
void sio_in_flush(void);
void sio_out_flush(void);
char _getkey(void);
char getchar(void);
char ungetchar(char character);
void putchar(char character);

bit sio_inrdy(void);
bit sio_inrdy_to_pack(void);
char sio_message();
bit sio_outrdy(void);
bit sio_outempty(void);

static void _sio_send(char character);
#if(CRC_CHECK)
unsigned int updcrc(unsigned char character, unsigned int oldcrc);
#endif

/***************************************************************************/
/*                                                                         */
/* void sio_init(char soft_flow_control)                                   */
/*                                                                         */
/* action: initialise both timer 1 and 2 and enable interrupts.            */
/*         baud rate and parity selection will be affected by jumper       */
/*         setting. you may select hard- or software flow control.         */
/*                                                                         */
/* jumper       state    result                                            */
/*                                                                         */
/* parity_sel       1     8 bit, no parity                                 */
/*                  0     7 bit, even parity                               */
/*                                                                         */
/* baud_sel1,0     00     9600 bd (baud_sel0 = lsb)                        */
/*                 01     4800 bd                                          */
/*                 10     2400 bd                                          */
/*                 11     1200 bd                                          */
/*                                                                         */
/***************************************************************************/

void sio_init(char soft_flow_control)
{
char x;
int code baud_table[] =
{/*   9615    4807    2403    1201 , with 12.000000MHz crystal */
 /* 0xffd9, 0xffb2, 0xff64, 0xfec8,                            */
 /*   9600    4800    2400    1200 , with 11.005920MHz crystal */
    0xffdc, 0xffb8, 0xff70, 0xfee0,
};

  even_parity = ~parity_sel;   /* read jumpers */
  x = baud_sel0 ? baud_sel1 ? 3 : 1 : baud_sel1 ? 2 : 0;
  hard_flow = soft_flow_control ? 0 : 1;

  TH2L2 = RCAP2H2L = baud_table[x];

  T2CON = 0x30;   /* RCLK, TCLK (TIMER 2) */
  SCON  = 0x72;   /* 8 BIT UART VAR., REN, TI */
  sio_in_flush();
  sio_out_flush();

#ifdef AUTO_XON
  give_xon_cnt = GIVE_XON;
#endif
#ifdef RELAXED_XOFF
  ignore_xoff_cnt = IGNORE_XOFF;
#endif

  ES = PS = TR2 = 1;

  TH1 = TIMER1_TIME >> 8;
  TL1 = TIMER1_TIME & 0xff;
  TMOD |= 0x10;
  ET1 = TR1 = 1;
}

/***************************************************************************/
/*                                                                         */
/* void sio_in_flush(void)                                                 */
/*                                                                         */
/* action: clear receive buffer and signal ready for receive               */
/*                                                                         */
/***************************************************************************/

void sio_in_flush(void)
{
char x;

  x = IE; EA = 0;

  rx_in = rx_out = rx_count = msg_count = 0;
  sent_xoff = send_xoff = parity_err = look_ahead_used = 0;
  send_xon = 1;

  IE = x;
}

/***************************************************************************/
/*                                                                         */
/* void sio_out_flush(void)                                                */
/*                                                                         */
/* action: clear transmit buffer                                           */
/*                                                                         */
/***************************************************************************/

void sio_out_flush(void)
{
char x;

  x = IE; EA = 0;

  tx_in = tx_out = tx_count = 0;
  tx_delay = CHAR_DELAY;
  tx_off = 0;

  IE = x;
}

/***************************************************************************/
/*                                                                         */
/* char getchar(void)                                                      */
/*                                                                         */
/* action: call _getkey() then echo character through putchar()            */
/*         and return the received character.                              */
/*         bit parity_err will be set on error.                            */
/*                                                                         */
/* this routine will serve your watchdog hardware during wait.             */
/*                                                                         */
/***************************************************************************/

char getchar(void)
{
char c;

  putchar(c = _getkey());
  return(c);
}

/***************************************************************************/
/*                                                                         */
/* char _getkey(void)                                                      */
/*                                                                         */
/* action: if a character has been stored due to ungetchar() then          */
/*         return this character else                                      */
/*         wait until receive buffer is not empty and                      */
/*         than return a character.                                        */
/*         bit parity_err will be set on error.                            */
/*                                                                         */
/* this routine will serve your watchdog hardware during wait.             */
/*                                                                         */
/***************************************************************************/

char _getkey(void)
{
char x, c;

  if (_testbit_(look_ahead_used))
    c = look_ahead;
  else
  {
    while(!sio_inrdy())
      ;

    x = IE; EA = 0;

    c = rx_buf[rx_out++];
    rx_out %= RX_BUF_LEN;

    if (--rx_count == 0 || rx_in == rx_out)
      rx_in = rx_out = msg_count = rx_count = 0;

    if (sent_xoff && rx_count < XON_LIMIT)
      send_xon = 1, sent_xoff = 0;
  }
  parity_err = 0;
  if (even_parity)
  {
    ACC = c;
    parity_err = P;                     /* odd if P is set */
  }

  if (!parity_err && c == TERMINATOR)
    --msg_count;

  IE = x;
  return(c);
}

/***************************************************************************/
/*                                                                         */
/* char ungetchar(char character)                                          */
/*                                                                         */
/* action: store given character and update msg_count                      */
/*                                                                         */
/***************************************************************************/

char ungetchar(char character)
{
  if (even_parity)
  {
    ACC = character;
    if (!P && ACC == TERMINATOR)
      ++msg_count;
  }
  look_ahead_used = 1;
  return(look_ahead = character);
}

/***************************************************************************/
/*                                                                         */
/* void putchar(char character)                                            */
/*                                                                         */
/* action: wait until transmit buffer can store a new character and        */
/*         than insert it into the buffer.                                 */
/*                                                                         */
/* this routine will serve your watchdog hardware during wait.             */
/*                                                                         */
/***************************************************************************/

void putchar(char character)
{
char x;

  while(!sio_outrdy())
    ;

  x = IE; EA = 0;

  if (even_parity)
  {
    ACC = character;
    ac7 = P;                            /* acc.7 */
  }

  tx_buf[tx_in++] = character;
  tx_in %= TX_BUF_LEN;
  tx_count++;

  if (tx_in == tx_out)                  /* only error correction */
    tx_count = TX_BUF_LEN;

  IE = x;
}

/***************************************************************************/
/*                                                                         */
/* bit sio_inrdy(void)                                                     */
/*                                                                         */
/* true, if there is a character in the receive buffer                     */
/*                                                                         */
/* you may use this routine in a loop because                              */
/* your watchdog hardware will be served.                                  */
/*                                                                         */
/***************************************************************************/

bit sio_inrdy(void)
{
#if(WATCHDOG)
  handle_watchdog();
#endif
  return rx_count > 0;
}

/***************************************************************************/
/*                                                                         */
/* bit sio_inrdy(void)                                                     */
/*                                                                         */
/* true, if there is a character in the receive buffer                     */
/*                                                                         */
/* you may use this routine in a loop because                              */
/* your watchdog hardware will be served.                                  */
/*                                                                         */
/***************************************************************************/

bit sio_inrdy_to_pack(void)
{
#if(WATCHDOG)
  handle_watchdog();
#endif
  return rx_count > 1;
}

/***************************************************************************/
/*                                                                         */
/* char sio_message(void)                                                  */
/*                                                                         */
/* returns the number of received TERMINATORS which may be defined         */
/* in sio_mod.h; if there are no message terminators in the receive buffer */
/* and a xoff has been send, then 0xff will be returned.                   */
/*                                                                         */
/* use this routine if you want to handle only full messages.              */
/* you may use this routine in a loop because                              */
/* your watchdog hardware will be served.                                  */
/*                                                                         */
/***************************************************************************/

char sio_message(void)
{
#if(WATCHDOG)
  handle_watchdog();
#endif

  if (!msg_count && sent_xoff)
    return 0xff;

  return msg_count;
}

/***************************************************************************/
/*                                                                         */
/* bit sio_outrdy(void)                                                    */
/*                                                                         */
/* true, if transmit buffer can take a new character                       */
/*                                                                         */
/* you may use this routine in a loop because                              */
/* your watchdog hardware will be served.                                  */
/*                                                                         */
/***************************************************************************/

bit sio_outrdy(void)
{
#if(WATCHDOG)
  handle_watchdog();
#endif
  return tx_count < TX_BUF_LEN;
}

/***************************************************************************/
/*                                                                         */
/* bit sio_outempty(void)                                                  */
/*                                                                         */
/* true, if transmit buffer is empty.                                      */
/*                                                                         */
/* call this routine if you need an empty buffer before you                */
/* start a transmission. you may use it in a loop because                  */
/* your watchdog hardware will be served.                                  */
/*                                                                         */
/***************************************************************************/

bit sio_outempty(void)
{
#if(WATCHDOG)
  handle_watchdog();
#endif
  return tx_count == 0;
}

/***************************************************************************/
/*                                                                         */
/* void sio_handler(void) interrupt SIO_INT                                */
/*                                                                         */
/* actions: handle received characters and flow control                    */
/*                                                                         */
/***************************************************************************/

static void sio_handler(void) interrupt SIO_INT using sio_bank
{
char character;

  if (_testbit_(TI))
    tx_ready = 1;

  if (_testbit_(RI))
  {
    character = SBUF;

    ACC = character;
    if (!P || !even_parity)             /* check only if no parity err */
    {                                   /* check without parity */
      if ((character &= 0x7f) == TERMINATOR)
        msg_count++;
      else
      {
        if (!hard_flow)
        {
          if (character == XON)
          {
            tx_off = 0;
            return;                     /* leave */
          }

          if (character == XOFF)
          {
            tx_off = 1;
            return;                     /* leave */
          }
        }
      }
    }

    if (rx_count < RX_BUF_LEN)
    {
      rx_buf[rx_in++] = character;
      rx_in %= RX_BUF_LEN;
      rx_count++;

      if (rx_in == rx_out)              /* only error correction */
        rx_count = RX_BUF_LEN;
    }

    if (rx_count >= XOFF_LIMIT)
      send_xoff = sent_xoff = 1;
  }
}

/***************************************************************************/
/*                                                                         */
/* void _sio_send(char character)                                          */
/*                                                                         */
/* used locally only                                                       */
/*                                                                         */
/***************************************************************************/

static void _sio_send(char character)
{
  tx_ready = 0;
  tx_delay = CHAR_DELAY;
  SBUF = character;
#ifdef AVSIM51
  TI = 1;
#endif
}

#if(CRC_CHECK)
/***************************************************************************/
/*                                                                         */
/* unsigned int updcrc(unsigned char character, unsigned int oldcrc)       */
/*                                                                         */
/* add character to oldcrc and return new crc                              */
/*                                                                         */
/***************************************************************************/

unsigned int updcrc(unsigned char character, unsigned int oldcrc)
{
int crc;
char x;

  crc = oldcrc;

  crc ^= (unsigned int)character;
  for(x=0; x<8; ++x)
  {
    if (crc & 1)
    {
      crc >>= 1;
      crc ^= 0x8408;                    /* CCITT - polynom */
    }
    else
      crc >>= 1;
  }
  return (crc);
}
#endif

/***************************************************************************/
/*                                                                         */
/* void timer1_handler(void) interrupt TIMER1_INT                          */
/*                                                                         */
/*     actions: handle various timeout counters used for communication     */
/*           and transmit characters out of tx_buf every CHAR_DELAY        */
/*                                                                         */
/***************************************************************************/

void timer1_handler(void) interrupt TIMER1_INT using sio_timer_bank
{
char c;

  /**************************************************************/
  /*** set new timer value with interrupt response correction ***/
  /**************************************************************/

union { word y; byte y1[2]; } x;

  TR1 = 0;
  x.y1[1] = TL1;
  x.y1[0] = TH1;
  x.y += TIMER1_TIME + 15;              /* this calculation lasts 15 us */
  TL1 = x.y1[1];
  TH1 = x.y1[0];
  TR1 = 1;

  /**************************************************************/

  if (hard_flow)
    tx_off = not_rts;                   /* enable/disable transmitter */

#ifdef AUTO_XON || RELAXED_XOFF
  else
  {
#endif
#ifdef AUTO_XON
    if (!msg_count)
    {
      if (--give_xon_cnt == 0)
      {
        give_xon_cnt = GIVE_XON;
        send_xon = 1;
      }
    }
    else
      give_xon_cnt = GIVE_XON;
#endif
#ifdef RELAXED_XOFF
    if (tx_off)
    {
      if (--ignore_xoff_cnt == 0)
      {
        ignore_xoff_cnt = IGNORE_XOFF;
        tx_off = 0;
      }
    }
    else
      ignore_xoff_cnt = IGNORE_XOFF;
#endif
#ifdef AUTO_XON || RELAXED_XOFF
  }
#endif

  if (tx_delay)                         /* handle delay counter */
    tx_delay--;

  if (tx_ready && !tx_delay)
  {
    if (_testbit_(send_xoff))
    {
      if (hard_flow)
        not_cts = 1;
      else
        _sio_send(XOFF);
    }
    else
    if (_testbit_(send_xon))
    {
      if (hard_flow)
        not_cts = 0;
      else
        _sio_send(XON);
    }
    else
    if (!tx_off && tx_count)
    {
      c = tx_buf[tx_out++];
      tx_out %= TX_BUF_LEN;

      if (--tx_count == 0 || tx_in == tx_out)
        tx_in = tx_out = tx_count = 0;

      _sio_send(c);
    }
  }
}

/***************************************************************************/

